local _G = getfenv(0);

-- Config Data Variables
TipTac_Config = {};
local cfg;

local TT_DefaultConfig = {
	point = "BOTTOMRIGHT",
	updateFreq = 0.5,
	tipScale = 1;
	gttScale = 1;
	preFadeTime = 0.8;
	fadeTime = 1.2;
	showTip = true;
	showTargetedBy = true;
	pvpName = true,
};

-- "Constants"
local TT_NotSpecified = "Not specified";
local TT_Classification = {
	["worldboss"] = "%s|r (Boss)",
	["rareelite"] = "+%s|r (Rare)",
	["elite"] = "+%s",
	["rare"] = "%s|r (Rare)",
};

-- Colors
local TT_WHITE = "|cffffffff";
local TT_LIGHTGRAY = "|cffc0c0c0";
local TT_GRAY = "|cff808080";
local TT_PURPLE = "|cffff32ff";
local TT_YELLOW = "|cffffff00";
local TT_LIGHTBLUE = "|cff25c1eb";

local TT_ClassColors = {};
for class, color in pairs(RAID_CLASS_COLORS) do
	TT_ClassColors[class] = ("|cff%.2x%.2x%.2x"):format(color.r*255,color.g*255,color.b*255);
end

local TT_ReactionColors = {};
for index, color in ipairs(UnitReactionColor) do
	tinsert(TT_ReactionColors,("|cff%.2x%.2x%.2x"):format(color.r*255,color.g*255,color.b*255));
end

local TT_DifficultyColors = {
	TT_GRAY,
	"|cff40c040",
	TT_YELLOW,
	"|cffff8040",
	"|cffff2020",
};

-- Data Variables
local u, w, player = {}, {};

--------------------------------------------------------------------------------------------------------
--                                           Frame Handlers                                           --
--------------------------------------------------------------------------------------------------------
function TipTacAnchor_OnLoad()
	-- Mod Info
	this.modName = "TipTac";
	this.modVers = GetAddOnMetadata(this.modName,"Version");
	-- Events
	this:RegisterEvent("UPDATE_MOUSEOVER_UNIT");
	this:RegisterEvent("PLAYER_LOGIN");
	this:RegisterEvent("PLAYER_LEVEL_UP");
	this:RegisterEvent("VARIABLES_LOADED");
	-- Add Slash Command
	SlashCmdList["TIPTAC"] = TipTac_OnSlash;
	SLASH_TIPTAC1 = "/tiptac";
	SLASH_TIPTAC2 = "/tip";
end

-- Events
function TipTacAnchor_OnEvent(event)
	-- Mouseover Changed
	if (event == "UPDATE_MOUSEOVER_UNIT") then
		TipTac_ShowUnit("mouseover");
	-- Player Login
	elseif (event == "PLAYER_LOGIN") then
		player = { name = UnitName("player"), level = UnitLevel("player") };
	-- Player Level Up
	elseif (event == "PLAYER_LEVEL_UP") then
		player.level = arg1;
	-- Variables Loaded
	elseif (event == "VARIABLES_LOADED") then
		cfg = TipTac_Config;
		for option, value in pairs(TT_DefaultConfig) do
			if (cfg[option] == nil) then
				cfg[option] = value;
			end
		end
		if (cfg.left and cfg.top) then
			TipTacAnchor:ClearAllPoints();
			TipTacAnchor:SetPoint("TOPLEFT",UIParent,"BOTTOMLEFT",cfg.left,cfg.top);
		end
		TipTac_ApplySettings();
	end
end

--------------------------------------------------------------------------------------------------------
--                                           Slash Handling                                           --
--------------------------------------------------------------------------------------------------------
function TipTac_OnSlash(cmd)
	-- Extract Parameters
	local param1, param2 = cmd:match("^([^%s]+)%s*(.*)$");
	param1 = (param1 and param1:lower() or cmd:lower());
	-- Options
	if (param1 == "") then
		local loaded, reason = LoadAddOn("TipTacOptions");
		if (loaded) then
			if (TipTacOptions:IsVisible()) then
				TipTacOptions:Hide();
			else
				TipTacOptions:Show();
			end
		else
			AzMsg("Could not option TicTac Options: |1"..tostring(reason));
		end
	-- Show Anchor
	elseif (param1 == "anchor") then
		TipTacAnchor:Show();
	-- DEBUG: GTT
	elseif (param1 == "gtt") then
		if (GameTooltip:NumLines() == 0) then
			AzMsg("GameTooltip has 0 lines.");
		else
			for i = 1, GameTooltip:NumLines() do
				AzMsg("|2"..i.."|r = |1"..(_G["GameTooltipTextLeft"..i]:GetText() or "").."   |2"..(_G["GameTooltipTextRight"..i]:GetText() or ""));
			end
		end
	-- Invalid or No Command
	else
		UpdateAddOnMemoryUsage();
		AzMsg("----- |2"..TipTacAnchor.modName.."|r |1"..TipTacAnchor.modVers.."|r ----- |1"..format("%.2f",GetAddOnMemoryUsage(TipTacAnchor.modName)).." |2kb|r -----");
		AzMsg("The following |2parameters|r are valid for |2/tiptac|r or |2/tip|r:");
		AzMsg(" |2anchor|r = Shows the anchor where the tooltip appears");
	end
end
--------------------------------------------------------------------------------------------------------
--                                Global Chat Message Function (Rev 2)                                --
--------------------------------------------------------------------------------------------------------
if (not AZMSG_REV or AZMSG_REV < 2) then
	AZMSG_REV = 2;
	function AzMsg(text)
		text = tostring(text):gsub("|1","|cffffff80"):gsub("|2","|cffffffff");
		DEFAULT_CHAT_FRAME:AddMessage(text,128/255,192/255,255/255);
	end
end
--------------------------------------------------------------------------------------------------------
--                                              TipTacTip                                             --
--------------------------------------------------------------------------------------------------------

-- Tip OnLoad
function TipTacTip_OnLoad()
	-- Backdrop
	--local backDrop = { bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", tile = true, tileSize = 1, edgeSize = 16, insets = { left = 4, right = 4, top = 4, bottom = 4 } };
	--local backDrop = { bgFile = "Interface\\AddOns\\AzMod\\MinimapMask", edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", tile = true, tileSize = 1, edgeSize = 16, insets = { left = 4, right = 4, top = 4, bottom = 4 } };
	--TipTacTip:SetBackdrop(backDrop);

	this.lastUpdate = 0;
	this:SetBackdropColor(0.1,0.1,0.2,1);
	this:SetBackdropBorderColor(0.3,0.3,0.4,1);
end

-- Tip OnUpdate
function TipTacTip_OnUpdate(elapsed)
	this.lastUpdate = (this.lastUpdate + elapsed);
	-- Fade Out
	if (this.fadeTime ~= -1) then
		if (GameTooltip:IsVisible()) then
			this:Hide();
		elseif (this.lastUpdate < cfg.preFadeTime) then
		elseif (this.lastUpdate - cfg.preFadeTime < this.fadeTime) then
			this:SetAlpha(1 - (this.lastUpdate - cfg.preFadeTime) / this.fadeTime);
		else
			this:Hide();
		end
	-- Hide Tip if unit is no more, hide GTT only if it is not showing anything new
	-- Az: Trying to fix this elseif so it will not mess up the hint going from a true mouseover to a mouseover on a unitframe
	--elseif (not UnitExists(u.token) or not UnitIsUnit(u.token,"mouseover") or u.name ~= UnitName("mouseover")) then
	elseif (not UnitExists(u.token) or not UnitIsUnit(u.token,"mouseover") or not GameTooltip:GetUnit() or GameTooltip:GetUnit() ~= UnitName(u.token)) then
		if (not GameTooltip:IsVisible() or GameTooltipTextLeft1:GetText() == u.gttName) then
			this.lastUpdate = 0;
			this.fadeTime = cfg.fadeTime;
			GameTooltip:Hide();
		else
			this:Hide();
		end
	-- Update Tip
	-- Az: Updating the GTT makes TipTac update extras like Beast Lore without moving the mouse off and on, more cpu used I guess?
	-- Although this sounds good, MobInfo2 updates the GTT in a very retarded way, so updating the tooltip with GTT:SetUnit() will
	-- not work as it clears the lines MobInfo2 added, and it doesn't add extra lines by hooking GTT:SetUnit().......
	elseif (cfg.updateFreq > 0) and (this.lastUpdate > cfg.updateFreq) then
		this.lastUpdate = 0;
		--GameTooltip:SetUnit(u.token);
		TipTac_ShowUnit(u.token);
	end

	-- Az: Debug
--[[	if (GameTooltip:IsVisible() and GameTooltip:GetAlpha() ~= 0 and TipTacTip:IsVisible()) then
		AzMsg("--- |2Both GTT and TTT Visible!|r ---")
		AzMsg("Cur GTT Line1 = |1"..GameTooltipTextLeft1:GetText():gsub("|","||"));
		AzMsg("Old GTT Line1 = |1"..u.gttName:gsub("|","||"));
		AzMsg("Cur TTT Line1 = |1"..TipTacTipTextLeft1:GetText():gsub("|","||"));
	end]]
end
--------------------------------------------------------------------------------------------------------
--                                               Hooks                                                --
--------------------------------------------------------------------------------------------------------

-- HOOK: GameTooltip_SetDefaultAnchor (For Re-Anchoring)
function GameTooltip_SetDefaultAnchor(tooltip,parent)
	if (cfg.point ~= "CURSOR") then
		tooltip:SetOwner(parent,"ANCHOR_NONE");
		tooltip:SetPoint(cfg.point,TipTacAnchor);
	else
		tooltip:SetOwner(parent,"ANCHOR_CURSOR");
	end
	tooltip.default = 1;
end

-- HOOK: GTT SetUnit (For Showing Units)
GameTooltip.SetUnit_TipTacOld = GameTooltip.SetUnit;
function GameTooltip:SetUnit(unit)
	self:SetUnit_TipTacOld(unit);
	TipTac_ShowUnit(unit);
end

-- HOOK: GTT OnShow (For Backdrop Color Change on Normal GTT)
GameTooltip.OnShow_TipTacOld = GameTooltip:GetScript("OnShow");
GameTooltip:SetScript("OnShow",
function()
	if (this.OnShow_TipTacOld) then
		this.OnShow_TipTacOld();
	end
	this:SetBackdropColor(0.1,0.1,0.2,1);		-- Def: For most: (0.1,0.1,0.2), for world objects?: (0,0.2,0.35)
	this:SetBackdropBorderColor(0.3,0.3,0.4,1);	-- Def: (1,1,1,1)
end
);

--------------------------------------------------------------------------------------------------------
--                                               Anchor                                               --
--------------------------------------------------------------------------------------------------------

local TT_AnchorPos = {
	{ name = "Top Right", point = "TOPRIGHT" },
	{ name = "Top Left", point = "TOPLEFT" },
	{ name = "Bottom Right", point = "BOTTOMRIGHT" },
	{ name = "Bottom Left", point = "BOTTOMLEFT" },
	{ name = "Top", point = "TOP" },
	{ name = "Bottom", point = "BOTTOM" },
	{ name = "Cursor", point = "CURSOR" },
};

-- Anchor OnMouseUp
function TipTacAnchor_OnMouseUp()
	if (arg1 == "LeftButton") then
		this:StopMovingOrSizing();
		cfg.left = TipTacAnchor:GetLeft();
		cfg.top = TipTacAnchor:GetTop();
	elseif (arg1 == "RightButton") then
		ToggleDropDownMenu(1,nil,TipTacAnchorDropDown,this:GetName(),0,0);
	end
end

-- DropDown OnLoad
function TipTacAnchor_DropDown_OnLoad()
	UIDropDownMenu_Initialize(this,TipTacAnchor_DropDown_Initialize,"MENU");
end

-- DropDown OnClick
function TipTacAnchor_DropDown_OnClick()
	cfg.point = this.value;
end

-- DropDown Init
local info = { func = TipTacAnchor_DropDown_OnClick };
function TipTacAnchor_DropDown_Initialize()
	for index, value in ipairs(TT_AnchorPos) do
		info.text = value.name;
		info.value = value.point;
		info.checked = (cfg and cfg.point == TT_AnchorPos[index].point);
		UIDropDownMenu_AddButton(info);
	end
end

--------------------------------------------------------------------------------------------------------
--                                           Tip Information                                          --
--------------------------------------------------------------------------------------------------------

-- Add Unit Details
function TipTac_AddUnitDetails()
	-- Reaction (Can't rely on UnitPlayerControlled() alone, since it always returns nil on units out of range)
	if (UnitIsPlayer(u.token) or UnitPlayerControlled(u.token)) then
		if (UnitCanAttack(u.token,"player")) then
			if (UnitCanAttack("player",u.token)) then
				u.reaction = TT_ReactionColors[2];
			else
				u.reaction = TT_ReactionColors[3];
			end
		elseif (UnitCanAttack("player",u.token)) then
			u.reaction = TT_ReactionColors[4];
		elseif (UnitIsPVP(u.token) and not UnitIsPVPSanctuary(u.token) and not UnitIsPVPSanctuary("player")) then
			u.reaction = TT_ReactionColors[6];
		else
			u.reaction = TT_LIGHTBLUE;
		end
	elseif (UnitIsTapped(u.token) and not UnitIsTappedByPlayer(u.token)) then
		u.reaction = TT_LIGHTGRAY;
	else
		u.reaction = TT_ReactionColors[UnitReaction(u.token,"player") or 5];
	end
	-- Players
	if (UnitIsPlayer(u.token)) then
		-- name
		u.name, u.realm = UnitName(u.token);
		u.line1 = u.reaction..(cfg.pvpName and UnitPVPName(u.token) or u.name);
		if (u.realm) then
			u.line1 = u.line1.." - "..u.realm;
		end
		-- afk or dnd
		if (UnitIsAFK(u.token)) then
			u.line1 = TT_WHITE.."<AFK> "..u.line1;
		elseif (UnitIsDND(u.token)) then
			u.line1 = TT_WHITE.."<DND> "..u.line1;
		end
		-- guild
		u.guild, u.guildRank = GetGuildInfo(u.token);
		if (u.guild) then
			player.guild = GetGuildInfo("player");
			if (u.guild == player.guild) then
				u.guild = TT_PURPLE.."<"..u.guild..">";
			else
				u.guild = u.reaction.."<"..u.guild..">";
			end
			--u.guild = u.guild.." "..TT_LIGHTGRAY..u.guildRank;
		end
		-- race
		u.race = " "..TT_WHITE..UnitRace(u.token);
		-- class
		u.class, u.classFixed = UnitClass(u.token);
		u.class = " "..(TT_ClassColors[u.classFixed] or TT_WHITE)..u.class;
	-- NPCs
	else
		-- name
		u.name = UnitName(u.token);
		u.line1 = u.reaction..u.name;
		-- guild
		u.guild = GameTooltipTextLeft2:GetText();
		u.guildRank = nil;
		if (u.guild) then
			if (u.guild:find("^"..LEVEL.." .+")) then
				u.guild = nil;
			else
				u.guild = u.reaction.."<"..u.guild..">";
			end
		end
		-- race
		u.race = "";
		-- class
		u.class = UnitCreatureFamily(u.token) or UnitCreatureType(u.token);
		if (not u.class or u.class == TT_NotSpecified) then
			u.class = "Unknown";
		end
		u.class = " "..TT_WHITE..u.class;
	end
	-- Target
	u.target = UnitName(u.token.."target");
	if (u.target == "Unknown") and (not UnitExists(u.token.."target")) then
		u.target = nil;
	elseif (u.target) then
		if (u.target == player.name) then
			u.target = "  "..TT_WHITE.."[YOU]"
		else
			u.target = "  "..(TT_ReactionColors[UnitReaction(u.token.."target","player")] or TT_LIGHTGRAY).."["..u.target.."]";
		end
	end
	-- Level + Classification
	u.level = UnitLevel(u.token);
	u.levelText = (u.level == -1 and "??" or u.level);
	u.classification = UnitClassification(u.token);
	if (TT_Classification[u.classification]) then
		u.levelText = TT_Classification[u.classification]:format(u.levelText);
	end
	if (UnitCanAttack(u.token,"player") or UnitCanAttack("player",u.token)) then
		u.levelText = TipTac_DifficultyLevel(u.level ~= -1 and u.level or 500)..u.levelText.."  ";
	else
		u.levelText = TT_LIGHTGRAY..u.levelText.." ";
	end
	-- Add The Stuff
	TipTacTip:AddLine(u.line1,1,1,1);
	if (u.target) then
		TipTacTip:AddLine(u.target,1,1,1);
		u.raidIcon = GetRaidTargetIndex(u.token.."target");
		if (u.raidIcon) then
			TipTacTip:AddTexture("Interface\\TargetingFrame\\UI-RaidTargetingIcons");
			SetRaidTargetIconTexture(TipTacTipTexture1,u.raidIcon);
		end
	end
	if (u.guild) then
		TipTacTip:AddLine(u.guild,1,1,1);
	end
	TipTacTip:AddLine(u.levelText..u.race..u.class,1,1,1);
end

-- Add Extra GTT Lines
function TipTac_AddExtraLines()
	for i = 2, GameTooltip:NumLines() do
		w.lineL = _G["GameTooltipTextLeft"..i]:GetText();
		if (w.lineL) and (w.lineL:find("^"..LEVEL.." .+")) then
			for i = i + 1, GameTooltip:NumLines() do
				w.lineL, w.lineR = _G["GameTooltipTextLeft"..i], _G["GameTooltipTextRight"..i];
				if (w.lineL:GetText() ~= PVP_ENABLED) then
					w.r, w.g, w.b = w.lineL:GetTextColor();
					TipTacTip:AddDoubleLine(w.lineL:GetText(),w.lineR:GetText(),w.r,w.g,w.b,w.lineR:GetTextColor());
				end
			end
			return;
		end
	end
end

-- Add "Targeted By" line
function TipTac_AddTargetedBy()
	w.numParty, w.numRaid = GetNumPartyMembers(), GetNumRaidMembers();
	if (w.numParty > 0 or w.numRaid > 0) then
		w.players, w.counter = "", 0;
		for i = 1, (w.numRaid > 0 and w.numRaid or w.numParty) do
			w.unit = (w.numRaid > 0 and "raid"..i or "party"..i);
			if (UnitIsUnit(w.unit.."target",u.token)) and (not UnitIsUnit(w.unit,"player")) then
				if (mod(w.counter + 3,6) == 0) then
					w.players = w.players.."\n";
				end
				w.color = RAID_CLASS_COLORS[select(2,UnitClass(w.unit))];
				w.color = ("|cff%.2x%.2x%.2x"):format(w.color.r*255,w.color.g*255,w.color.b*255);
				w.players = w.players..w.color..UnitName(w.unit).."|r, ";
				w.counter = (w.counter + 1);
			end
		end
		if (w.players ~= "") then
			TipTacTip:AddLine("Targeted By (|cffffffff"..w.counter.."|r): "..w.players:sub(1,-5));
		end
	end
end

-- Update Health & Power
function TipTac_UpdateHealthAndPowerBar()
	-- Health
	if (TipTacTipHealth:IsShown()) then
		u.hp, u.hpMax = UnitHealth(u.token), UnitHealthMax(u.token);
		TipTacTipHealth:SetMinMaxValues(0,u.hpMax);
		TipTacTipHealth:SetValue(u.hp);
		if (u.hpMax ~= 100 or UnitIsUnit(u.token,"player") or UnitIsUnit(u.token,"pet") or UnitPlayerOrPetInParty(u.token) or UnitPlayerOrPetInRaid(u.token))  then
			TipTacTipHealthText:SetText(u.hp.." / "..u.hpMax);
		elseif (MobHealth_PPP) then
			u.ppp = MobHealth_PPP(u.name..":"..u.level);
			if (u.ppp) and (u.ppp > 0) then
				u.hp, u.hpMax = floor(u.ppp * u.hp + 0.5), floor(u.ppp * 100 + 0.5);
				TipTacTipHealthText:SetText(u.hp.." / "..u.hpMax);
			else
				TipTacTipHealthText:SetText(u.hp.."%");
			end
		else
			TipTacTipHealthText:SetText(u.hp.."%");
		end
	end
	-- Power
	--[[if (TipTacTipPower:IsShown()) then
		u.power, u.powerMax = UnitMana(u.token), UnitManaMax(u.token);
		TipTacTipPower:SetMinMaxValues(0,u.powerMax);
		TipTacTipPower:SetValue(u.power);
		--TipTacTipPowerText:SetText(u.power.." / "..u.powerMax);
		local color = ManaBarColor[UnitPowerType(u.token)] or { r = 0.5, g= 0.5, b = 0.5 };
		TipTacTipPower:SetStatusBarColor(color.r,color.g,color.b);
	end]]
	-- Cast Bar
	--[[if (1) then
		local spell, rank, displayName, icon, startTime, endTime = UnitCastingInfo(u.token);
		if (not spell) then
			spell, rank, displayName, icon, startTime, endTime = UnitChannelInfo(u.token);
			if (not spell) then
				return;
			end
		end
		startTime, endTime = startTime/1000, endTime/1000;
		TipTacTipHealth:SetMinMaxValues(startTime,endTime);
		TipTacTipHealth:SetValue(GetTime());
		TipTacTipHealthText:SetText(format("%s %d%%",spell,((GetTime()-startTime) / (endTime-startTime) * 100)));
	end]]
end
--------------------------------------------------------------------------------------------------------
--                                                Misc                                                --
--------------------------------------------------------------------------------------------------------

-- Apply Settings
function TipTac_ApplySettings()
	TipTacTip:SetScale(cfg.tipScale);
	GameTooltip:SetScale(cfg.gttScale);
end

-- Returns the correct difficulty color compared to the player
function TipTac_DifficultyLevel(level)
	level = (level - player.level);
	if (level > 4) then
		return TT_DifficultyColors[5]; -- red
	elseif (level > 2) then
		return TT_DifficultyColors[4]; -- orange
	elseif (level >= -2) then
		return TT_DifficultyColors[3]; -- yellow
	elseif (level >= -GetQuestGreenRange()) then
		return TT_DifficultyColors[2]; -- green
	else
		return TT_DifficultyColors[1]; -- gray
	end
end

-- Show Unit
function TipTac_ShowUnit(unitToken)
	if (not cfg.showTip or not UnitExists(unitToken)) then
		return;
	end
	-- Init
	u.token = unitToken;
	u.gttName = GameTooltipTextLeft1:GetText();
	GameTooltip_SetDefaultAnchor(TipTacTip,UIParent);
	-- Construct the TipTacTip
	TipTac_AddUnitDetails();
	TipTac_AddExtraLines();
	if (cfg.showTargetedBy) then
		TipTac_AddTargetedBy();
	end
	TipTac_UpdateHealthAndPowerBar();
	-- Hide GTT and Show TTT
	GameTooltip:SetAlpha(0);
	TipTacTip.fadeTime = -1;
	TipTacTip:Show();
	-- Adjust Height
	TipTacTip:SetHeight(TipTacTip:GetHeight() + 11);
end