local _G = getfenv(0);

local cfg = TipTac_Config;

local TTO_Sliders = {
	{ var = "updateFreq", label = "Tip Update Frequency", min = 0, max = 5, step = 0.05 },
	{ var = "tipScale", label = "TipTac Scale", min = 0.4, max = 2, step = 0.05 },
	{ var = "gttScale", label = "GameTooltip Scale", min = 0.4, max = 2, step = 0.05 },
	{ var = "preFadeTime", label = "Pre Fade Time", min = 0, max = 5, step = 0.05 },
	{ var = "fadeTime", label = "Fade Out Time", min = 0, max = 5, step = 0.05 },
};

local TTO_Checks = {
	{ var = "showTip", label = "Show TipTac Unit Tips", tip = "If you disable this option, you will only really take advantage of the anchor option in TipTac" },
	{ var = "showTargetedBy", label = "Show Who Targets the Unit", tip = "When in a raid or party, it will show who from your group is targeting the unit" },
	{ var = "pvpName", label = "Use PvP Name If Available", tip = "Show PvP title for players if available" },
};

--------------------------------------------------------------------------------------------------------
--                                           Frame Scripts                                            --
--------------------------------------------------------------------------------------------------------
function TipTacOptions_OnLoad()
	TipTacOptionsVers:SetText("|cffffff80"..TipTacAnchor.modVers);
	if (TipTacAnchor) then
		TipTacOptionsAnchor:Show();
	end
end

--------------------------------------------------------------------------------------------------------
--                                            Slider Frame                                            --
--------------------------------------------------------------------------------------------------------

-- Slider Frame OnLoad
function TipTacOptions_Slider_OnLoad()
	local id = this:GetID();
	local slider = (this:GetName().."Slider");

	_G[slider]:SetMinMaxValues(TTO_Sliders[id].min,TTO_Sliders[id].max);
	_G[slider]:SetValueStep(TTO_Sliders[id].step);
	_G[slider]:SetValue(cfg[TTO_Sliders[id].var]);

	_G[slider.."Text"]:SetText(TTO_Sliders[id].label);
	_G[slider.."Low"]:SetText(TTO_Sliders[id].min);
	_G[slider.."High"]:SetText(TTO_Sliders[id].max);
end

-- Slider Value Changed
function TipTacOptions_Slider_OnValueChanged()
	local id = this:GetParent():GetID();
	-- Tooltip
	if (GameTooltip:IsOwned(this)) then
		GameTooltip:SetText(format("%.2f",this:GetValue()));
	end
	-- Change Edit
	_G[this:GetParent():GetName().."Edit"]:SetNumber(this:GetValue());
	-- Save
	cfg[TTO_Sliders[id].var] = this:GetValue();
	-- Apply
	TipTac_ApplySettings();
end

-- EditBox Enter Pressed
function TipTacOptions_SliderEdit_OnEnterPressed()
	_G[this:GetParent():GetName().."Slider"]:SetValue(this:GetNumber());
end

--------------------------------------------------------------------------------------------------------
--                                           Check Buttons                                            --
--------------------------------------------------------------------------------------------------------
function TipTacOptions_CheckBox_OnLoad()
	local id = this:GetID();
	_G[this:GetName().."Text"]:SetText(TTO_Checks[id].label);
	this.tooltipText = TTO_Checks[id].tip;
	this:SetChecked(cfg[TTO_Checks[id].var]);
end

function TipTacOptions_CheckBox_OnClick()
	cfg[TTO_Checks[this:GetID()].var] = (this:GetChecked() == 1);
end

--------------------------------------------------------------------------------------------------------
--                                                Misc                                                --
--------------------------------------------------------------------------------------------------------
function TipTacOptions_Anchor()
	if (TipTacAnchor:IsVisible()) then
		TipTacAnchor:Hide();
	else
		TipTacAnchor:Show();
	end
end